package info.textgrid.lab.collatex.ui.parts;

import info.textgrid.lab.collatex.CollateXPlugin;
import info.textgrid.lab.collatex.model.CollationResults;
import info.textgrid.lab.collatex.model.CollationSet;
import info.textgrid.lab.collatex.model.CollationSetWitness;
import info.textgrid.lab.collatex.process.CollateJob;
import info.textgrid.lab.collatex.ui.CollateXPerspective;
import info.textgrid.lab.collatex.ui.utils.JobWithUICleanup;
import info.textgrid.lab.collatex.ui.utils.NotificationViewerElement;
import info.textgrid.lab.collatex.ui.utils.SaveConfirmation;
import info.textgrid.lab.collatex.ui.utils.TGONotificationAwareLabelProvider;
import info.textgrid.lab.core.model.CrudServiceException;
import info.textgrid.lab.core.model.TGContentType;
import info.textgrid.lab.core.model.TextGridObject;
import info.textgrid.lab.core.swtutils.AdapterUtils;
import info.textgrid.lab.core.swtutils.CoreExceptionStatus;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.net.URI;
import java.util.Iterator;
import java.util.List;

import javax.xml.bind.DataBindingException;
import javax.xml.bind.JAXB;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IStorage;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.AssertionFailedException;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.ListenerList;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.util.LocalSelectionTransfer;
import org.eclipse.jface.viewers.CheckStateChangedEvent;
import org.eclipse.jface.viewers.CheckboxTreeViewer;
import org.eclipse.jface.viewers.ICheckStateListener;
import org.eclipse.jface.viewers.ICheckStateProvider;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerDropAdapter;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DragSourceEvent;
import org.eclipse.swt.dnd.DropTarget;
import org.eclipse.swt.dnd.DropTargetEffect;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.dnd.TransferData;
import org.eclipse.swt.dnd.TreeDragSourceEffect;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Link;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.TreeColumn;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.IReusableEditor;
import org.eclipse.ui.IStorageEditorInput;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.WorkbenchException;
import org.eclipse.ui.contexts.IContextService;
import org.eclipse.ui.dialogs.ContainerCheckedTreeViewer;
import org.eclipse.ui.part.EditorPart;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.progress.UIJob;
import org.eclipse.ui.statushandlers.StatusManager;
import org.eclipse.wb.swt.ResourceManager;

import com.google.common.base.Predicate;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;

/**
 * Editor for a collation set. The active {@link CollationSetEditor} also
 * determines which {@linkplain AbstractResultPage result pages} will be shown,
 * and the editor manages not only the {@link CollationSet}, but also the
 * {@link CollationResults}.
 */
@SuppressWarnings("restriction")
public class CollationSetEditor extends EditorPart implements IReusableEditor, IResourceChangeListener {

	/**
	 * Event listener that is notified when a collation run has finished for
	 * this editor and thus new {@link CollationResults} object is available.
	 */
	public interface ICollationEventHandler {
		public void newCollationResults(final CollationResults collationResults);
	}

	private final ListenerList collationEventHandlers = new ListenerList();

	public void addCollationListener(final ICollationEventHandler listener) {
		collationEventHandlers.add(listener);
	}

	public void removeCollationListener(final ICollationEventHandler listener) {
		collationEventHandlers.remove(listener);
	}

	/**
	 * Drop adapter for the link to the normalization spec.
	 */
	private final class RegularisationDropAdapter extends
	DropTargetEffect {
		private RegularisationDropAdapter(final Control control) {
			super(control);
		}

		@Override
		public void drop(final DropTargetEvent event) {
			if (accept(event)) {
				final TextGridObject equivsTGO = getEquivsTGO();
				collationSet.setRegularisation(equivsTGO.getPreparedURI());
				setDirty(true);
				((Link) getControl()).setText(NLS.bind(
						Messages.CollationSetEditor_RegularisationLink, equivsTGO));
			}
		}

		/**
		 * @return the {@link TextGridObject} that corresponds to the currently
		 *         chosen normalisation (or <code>null</code> if none)
		 */
		private TextGridObject getEquivsTGO() {
			final ISelection selection = LocalSelectionTransfer
					.getTransfer().getSelection();
			if (selection instanceof IStructuredSelection) {
				final IStructuredSelection sel = (IStructuredSelection) selection;
				final TextGridObject textGridObject = AdapterUtils
						.getAdapter(sel.getFirstElement(),
								TextGridObject.class);
				try {
					if (textGridObject != null
							&& CollationSetEditor.REGULARISATION_TYPE
							.equals(textGridObject
									.getContentType(false)))
						return textGridObject;
					else
						return null;
				} catch (final IllegalArgumentException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (final CoreException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			return null;
		}

		@Override
		public void dropAccept(final DropTargetEvent event) {
			if (accept(event))
				event.operations = DND.DROP_NONE;
		}

		private boolean accept(final DropTargetEvent event) {
			boolean accept = false;
			for (final TransferData type : event.dataTypes)
				accept |= LocalSelectionTransfer.getTransfer()
				.isSupportedType(type);
			return accept && getEquivsTGO() != null;
		}
	}

	/**
	 * Drop adapter for the list of witnesses.
	 */
	private final class CollationSetDropAdapter extends ViewerDropAdapter {

		private CollationSetDropAdapter(final Viewer viewer) {
			super(viewer);
			setFeedbackEnabled(true);
		}

		@Override
		public boolean validateDrop(final Object target, final int operation,
				final TransferData transferType) {

			final boolean supportedType = LocalSelectionTransfer.getTransfer()
					.isSupportedType(transferType);
			// System.out.println(MessageFormat.format(
			//					"CollationSetDropAdapter: Drop {0} on {1}, type {2}? {3}", //$NON-NLS-1$
			// target,
			// LocalSelectionTransfer.getTransfer().getSelection(),
			//					transferType, supportedType ? "yes" : "no")); //$NON-NLS-1$ //$NON-NLS-2$

			return supportedType && getCurrentLocation() != LOCATION_ON;
		}

		@Override
		public boolean performDrop(final Object data) {
			boolean result = false;
			final LocalSelectionTransfer transfer = LocalSelectionTransfer
					.getTransfer();
			if (transfer.isSupportedType(getCurrentEvent().currentDataType)) {
				final ISelection selection = transfer.getSelection();
				if (selection instanceof IStructuredSelection) {
					final Iterator<?> iterator = ((IStructuredSelection) selection)
							.iterator();
					final List<CollationSetWitness> newWitnesses = Lists
							.newLinkedList();
					while (iterator.hasNext()) {
						final Object next = iterator.next();
						if (next instanceof CollationSetWitness) {
							newWitnesses.add((CollationSetWitness) next);
							result = true;
						} else {
							final TextGridObject tgo = AdapterUtils.getAdapter(
									next, TextGridObject.class);
							if (tgo != null) {
								newWitnesses.add(new CollationSetWitness(tgo
										.getURI()));
								result = true;
							} else
								StatusManager.getManager().handle(new Status(IStatus.WARNING, CollateXPlugin.PLUGIN_ID, NLS.bind("Cannot make a witness out if {0} (a {1})", next, next.getClass().getSimpleName()))); //$NON-NLS-1$
						}

					}
					// System.out.println(newWitnesses);
					if (!newWitnesses.isEmpty()) {
						// remove old instances of moved items
						collationSet.getWitnesses().removeAll(newWitnesses);

						switch(getCurrentLocation()) {
						case LOCATION_AFTER:
						case LOCATION_ON:
							collationSet.getWitnesses().addAll(collationSet.getWitnesses().indexOf(getCurrentTarget())+1, newWitnesses);
							break;
						case LOCATION_BEFORE:
							collationSet.getWitnesses().addAll(collationSet.getWitnesses().indexOf(getCurrentTarget()), newWitnesses);
							break;
						case LOCATION_NONE:
							collationSet.getWitnesses().addAll(newWitnesses);

						}
						setDirty(true);
						//						System.out.println("Added: " //$NON-NLS-1$
						//								+ Joiner.on(", ").join(newWitnesses)); //$NON-NLS-1$
						getViewer().refresh();
					}
				}
			}

			return result;
		}
	}

	/**
	 * Content provider that delivers {@link CollationSetWitness}es from a
	 * {@link CollationSet}. Also handles the
	 * {@link CollationSetWitness#isActive()} active state.
	 * 
	 * <p>
	 * Supports not-yet-loaded collation sets (i.e. viewer input is
	 * <code>null</code>) and empty collation sets by providing appropriate
	 * labels by means of a {@link NotificationViewerElement}.
	 */
	private final class CollationSetContentProvider implements
	ITreeContentProvider, ICheckStateProvider {
		private CollationSet input;

		@Override
		public void dispose() {
			input = null;
		}

		@Override
		public void inputChanged(final Viewer viewer,
				final Object oldInput, final Object newInput) {
			if (newInput instanceof CollationSet)
				input = (CollationSet) newInput;
			// else
			// throw new IllegalArgumentException(MessageFormat.format(
			//						"{0} should be a CollationSet", newInput)); //$NON-NLS-1$
		}

		@Override
		public Object[] getElements(final Object inputElement) {
			if (inputElement == null)
				return new Object[] { new NotificationViewerElement(
						Messages.CollationSetEditor_Loading) };
			if (inputElement instanceof CollationSet)
				if (((CollationSet) inputElement).getWitnesses().isEmpty())
					return new Object[] { new NotificationViewerElement(
							Messages.CollationSetEditor_DropWitnessesHere) };
				else
					return ((CollationSet) inputElement).getWitnesses()
							.toArray();
			else
				return new Object[0];
		}

		@Override
		public Object[] getChildren(final Object parentElement) {
			return getElements(parentElement);
		}

		@Override
		public Object getParent(final Object element) {
			if (input.getWitnesses().contains(element))
				return input;
			else
				return null;
		}

		@Override
		public boolean hasChildren(final Object element) {
			return (element instanceof CollationSet && !((CollationSet)element).getWitnesses().isEmpty());
		}

		@Override
		public boolean isChecked(final Object element) {
			if (element instanceof CollationSetWitness)
				return ((CollationSetWitness) element).isActive();
			else
				return false;
		}

		@Override
		public boolean isGrayed(final Object element) {
			return false;
		}
	}

	public static final String ID = "info.textgrid.lab.collatex.collationSetEditor"; //$NON-NLS-1$
	private CheckboxTreeViewer collationSetViewer;
	private CollationSet collationSet;

	public CollationSet getCollationSet() {
		return collationSet;
	}

	private boolean dirty;
	private static final TGContentType REGULARISATION_TYPE = TGContentType
			.of("text/collation-equivs+xml"); //$NON-NLS-1$
	private CollateJob collateJob;
	private CollationResults collationResults;
	private IFile file;
	protected boolean readOnly;
	private Link regularisationLink;

	public CollationResults getCollationResults() {
		return collationResults;
	}

	/**
	 * Called after the {@link CollateJob} is done. May be called from outside
	 * of the UI thread.
	 */
	protected void setCollationResults(final CollationResults collationResults) {
		this.collationResults = collationResults;

		new UIJob("") { //$NON-NLS-1$

			@Override
			public IStatus runInUIThread(final IProgressMonitor monitor) {
				for (final Object listener : collationEventHandlers.getListeners())
					((ICollationEventHandler) listener)
					.newCollationResults(collationResults);
				return Status.OK_STATUS;
			}
		}.schedule();
	}
	public CollationSetEditor() {
	}

	/**
	 * Create contents of the editor part.
	 * 
	 * @param parent
	 */
	@SuppressWarnings("restriction")
	@Override
	public void createPartControl(final Composite parent) {
		final Composite container = new Composite(parent, SWT.NONE);
		container.setLayout(new GridLayout(1, false));

		collationSetViewer = new ContainerCheckedTreeViewer(container,
				SWT.BORDER | SWT.CHECK | SWT.MULTI);
		final Tree tree = collationSetViewer.getTree();
		tree.setHeaderVisible(true);
		tree.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1));

		final CollationSetContentProvider contentProvider = new CollationSetContentProvider();
		collationSetViewer.setContentProvider(contentProvider);
		collationSetViewer.setCheckStateProvider(contentProvider);

		final TreeViewerColumn titleColumn = new TreeViewerColumn(
				collationSetViewer, SWT.NONE);
		final TreeColumn titleTreeColumn = titleColumn.getColumn();
		titleTreeColumn.setWidth(200);
		titleTreeColumn.setText(Messages.CollationSetEditor_WitnessTitle);

		titleColumn.setLabelProvider(new TGONotificationAwareLabelProvider());

		tree.addControlListener(new ControlAdapter() {
			@Override
			public void controlResized(final ControlEvent e) {
				titleTreeColumn.setWidth(tree.getBounds().width - 2
						* tree.getBorderWidth());
			}
		});

		collationSetViewer.addDropSupport(DND.DROP_MOVE | DND.DROP_DEFAULT,
				new Transfer[] { LocalSelectionTransfer.getTransfer() },
				new CollationSetDropAdapter(collationSetViewer));

		collationSetViewer.addDragSupport(DND.DROP_MOVE | DND.DROP_DEFAULT,
				new Transfer[] { LocalSelectionTransfer.getTransfer() },
				new TreeDragSourceEffect(tree) {

			@Override
			public void dragStart(final DragSourceEvent event) {
				super.dragStart(event);
				ISelection selection = collationSetViewer
						.getSelection();
				if (selection.isEmpty())
					selection = new StructuredSelection(event
							.getSource());
				LocalSelectionTransfer.getTransfer().setSelection(
						selection);
			}

			@Override
			public void dragFinished(final DragSourceEvent event) {
				super.dragFinished(event);
				//				final ISelection selection = LocalSelectionTransfer
				//						.getTransfer().getSelection();
				//				if (selection instanceof IStructuredSelection) {
				//					final IStructuredSelection sel = (IStructuredSelection) selection;
				//					final CollationSetWitness[] witnesses = AdapterUtils
				//							.getAdapters(sel.toArray(),
				//									CollationSetWitness.class, false);
				//					collationSet.getWitnesses().removeAll(
				//							Arrays.asList(witnesses));
				//							System.out.println("Removed "
				//									+ Joiner.on(", ").join(witnesses));
				//					collationSetViewer.remove(witnesses);
				//					// collationSetViewer.refresh();
				// }
			}

		});
		collationSetViewer.addCheckStateListener(new ICheckStateListener() {
			@Override
			public void checkStateChanged(final CheckStateChangedEvent event) {
				if (event.getElement() instanceof CollationSetWitness)
					((CollationSetWitness) event.getElement()).setActive(event
							.getChecked());
			}
		});

		collationSetViewer
		.addSelectionChangedListener(new ISelectionChangedListener() {

			@Override
			public void selectionChanged(
					final SelectionChangedEvent event) {
				final IStructuredSelection sel = (IStructuredSelection) event
						.getSelection();
				if (!sel.isEmpty()
						&& sel.getFirstElement() instanceof CollationSetWitness
						&& collationResults != null)
					collationResults
					.setFocusedCSWitness((CollationSetWitness) sel
							.getFirstElement());
			}
		});

		collationSetViewer.setInput(collationSet);

		getSite().setSelectionProvider(collationSetViewer);

		createRegularizationArea(container);

		getEditorSite().setSelectionProvider(collationSetViewer);

		// final Control control = ((PartSite)
		// getSite()).getPane().getControl();
		PlatformUI
		.getWorkbench()
		.getHelpSystem()
				.setHelp(container,
				"info.textgrid.lab.collatex.collationSetEditor"); //$NON-NLS-1$
	}

	private void createRegularizationArea(final Composite container) {
		final Composite composite = new Composite(container, SWT.BORDER);
		composite.setLayout(new GridLayout(3, false));
		composite.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false,
				1, 1));

		regularisationLink = new Link(composite, SWT.NONE);
		regularisationLink.setLayoutData(new GridData(SWT.FILL, SWT.CENTER,
				true, false, 1, 1));
		if (collationSet == null || collationSet.getRegularisation() == null)
			regularisationLink
			.setText(Messages.CollationSetEditor_NewRegularisationLink);
		else {
			final TextGridObject regObject = TextGridObject
					.getInstanceOffline(collationSet.getRegularisation());
			regularisationLink.setText(NLS.bind(Messages.CollationSetEditor_RegularisationLink,
					regObject));
		}

		final DropTarget regularisationDropTarget = new DropTarget(
				regularisationLink, DND.DROP_MOVE | DND.DROP_COPY
				| DND.DROP_LINK);
		regularisationDropTarget
		.setTransfer(new Transfer[] { LocalSelectionTransfer
				.getTransfer() });
		regularisationDropTarget.addDropListener(new RegularisationDropAdapter(regularisationLink));

		final Button browseForRegularisationButton = new Button(composite,
				SWT.FLAT);
		browseForRegularisationButton
		.setToolTipText(Messages.CollationSetEditor_browseForRegularisationButton_toolTipText);
		browseForRegularisationButton
		.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(final SelectionEvent e) {
				final TextGridObject newEquivalenceSet = SelectObjectDialog
						.selectObject(
								getSite().getShell(),
								"format:\"text/collation-equivs+xml\"", //$NON-NLS-1$
								Messages.CollationSetEditor_SelectEquivalenceSetTitle,
								Messages.CollationSetEditor_SelectEquivalenceSetTitle,
								Messages.CollationSetEditor_SelectEquivalenceSetMessage,
								null);
				if (newEquivalenceSet != null) {
					collationSet.setRegularisation(URI
							.create(newEquivalenceSet.getLatestURI()));
					setDirty(true);
					regularisationLink.setText(NLS
							.bind(Messages.CollationSetEditor_RegularisationLink,
									newEquivalenceSet));
				}

			}
		});
		browseForRegularisationButton.setImage(ResourceManager.getPluginImage(
				"info.textgrid.lab.collatex", "resources/browsebutton.gif")); //$NON-NLS-1$ //$NON-NLS-2$
		browseForRegularisationButton.setLayoutData(new GridData(SWT.CENTER,
				SWT.FILL, false, false, 1, 1));

		final Button clearRegularisationButton = new Button(composite, SWT.FLAT
				| SWT.CENTER);
		clearRegularisationButton
		.setToolTipText(Messages.CollationSetEditor_clearRegularisationButton_toolTipText);
		clearRegularisationButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(final SelectionEvent e) {
				collationSet.setRegularisation(null);
				setDirty(true);
				regularisationLink
				.setText(Messages.CollationSetEditor_NewRegularisationLink);
			}
		});
		clearRegularisationButton.setLayoutData(new GridData(SWT.CENTER,
				SWT.FILL, false, false, 1, 1));
		clearRegularisationButton.setImage(ResourceManager.getPluginImage(
				"info.textgrid.lab.collatex", "resources/delete.gif")); //$NON-NLS-1$ //$NON-NLS-2$
		regularisationLink.addSelectionListener(new SelectionListener() {


			@Override
			public void widgetSelected(final SelectionEvent e) {
				if (collationSet.getRegularisation() == null) {
					// TODO move to background job
					if (getInputObject() != null)
						createRegularisationSet();
				} else {
					final IFileEditorInput input = AdapterUtils
							.getAdapter(
									TextGridObject.getInstanceOffline(collationSet
											.getRegularisation()),
											IFileEditorInput.class);
					try {
						getSite().getPage().openEditor(input,
								RegularisationEditor.ID, true);
					} catch (final PartInitException e1) {
						StatusManager.getManager().handle(e1,
								CollateXPlugin.PLUGIN_ID);
					}

				}

			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e) {
			}
		});
	}

	/**
	 * Helper method that creates a new normalization and associates it with the
	 * {@link CollationSet} in this editor. Note that this immediately requests
	 * an URI.
	 */
	public void createRegularisationSet() {
		try {
			final TextGridObject regObject = TextGridObject.getNewObjectInstance(getInputObject().getProjectInstance(), REGULARISATION_TYPE);
			TextGridObject.fetchURIs(ImmutableList.of(regObject), null);
			collationSet.setRegularisation(regObject
					.getPreparedURI());
			regularisationLink.setText(NLS.bind(
					Messages.CollationSetEditor_RegularisationLink, regObject));
			setDirty(true);
			regObject.setTitle(NLS.bind(
					Messages.CollationSetEditor_RegularisationDefaultTitle, getInputObject()
					.getTitle()));
			final IFileEditorInput input = AdapterUtils
					.getAdapter(regObject,
							IFileEditorInput.class);
			getSite().getPage().openEditor(input,
					RegularisationEditor.ID, true);
		} catch (final CoreException e1) {
			StatusManager.getManager().handle(e1,
					CollateXPlugin.PLUGIN_ID);
		}
	}


	@Override
	public void setFocus() {
		// Set the focus
	}

	@Override
	public void doSave(final IProgressMonitor monitor) {
		final SubMonitor progress = SubMonitor.convert(monitor,
				NLS.bind(Messages.CollationSetEditor_Saving_x, getEditorInput().getName()), 100);
		if (progress.isCanceled())
			return;
		final ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		JAXB.marshal(collationSet, outputStream);
		progress.worked(15);
		final IFile file = ((IFileEditorInput) getEditorInput()).getFile();
		try {
			file.setContents(
					new ByteArrayInputStream(outputStream.toByteArray()), true,
					true, progress.newChild(80));
			if (progress.isCanceled())
				return;
			setDirty(false);
			progress.worked(5);
			progress.done();
		} catch (final CoreException e) {
			StatusManager.getManager().handle(
					new Status(IStatus.ERROR, CollateXPlugin.PLUGIN_ID,
							NLS.bind(Messages.CollationSetEditor_Could_not_save_x, getEditorInput()
									.getName()), e),
									StatusManager.SHOW | StatusManager.LOG);
			progress.setCanceled(true);
		}
	}

	/**
	 * Not yet implemented.
	 * 
	 * @see org.eclipse.ui.part.EditorPart#doSaveAs()
	 */
	@Override
	public void doSaveAs() {
		// Do the Save As operation
	}

	/**
	 * Loading happens in the background!
	 * 
	 * @see org.eclipse.ui.part.EditorPart#init(org.eclipse.ui.IEditorSite,
	 *      org.eclipse.ui.IEditorInput)
	 * @see #setInput(IEditorInput)
	 */
	@Override
	public void init(final IEditorSite site, final IEditorInput input)
			throws PartInitException {

		setSite(site);
		ResourcesPlugin.getWorkspace().addResourceChangeListener(this, IResourceChangeEvent.POST_CHANGE);

		if (input instanceof IStorageEditorInput)
			setInput(input);
		else throw new PartInitException(NLS.bind(Messages.CollationSetEditor_InvalidEditorInput_x, input));


		final IContextService contextService = (IContextService) site
				.getService(IContextService.class);
		contextService
		.activateContext("info.textgrid.lab.collatex.editingCollationSets"); //$NON-NLS-1$

		try {
			PlatformUI.getWorkbench().showPerspective(CollateXPerspective.ID,
					site.getWorkbenchWindow());
		} catch (final WorkbenchException e) {
			StatusManager.getManager().handle(e, CollateXPlugin.PLUGIN_ID);
		}
	}

	/**
	 * Causes the given input to be loaded in the background. Fires an
	 * appropriate notification when the input has been completely loaded.
	 * 
	 * @see org.eclipse.ui.part.EditorPart#setInput(org.eclipse.ui.IEditorInput)
	 */
	@Override
	public void setInput(final IEditorInput input) {
		/*
		 * Regarding the input, there are three valid options: 1. Input is an
		 * existing file w/collation set -> it is parsed, success -> ok 2. Input
		 * is a non-existing file -> new collation set
		 */
		if (input instanceof IStorageEditorInput) {
			super.setInput(input);
			if (input instanceof IFileEditorInput)
				this.file = ((IFileEditorInput) input).getFile();
			else
				this.file = null;

			new JobWithUICleanup(Messages.CollationSetEditor_LoadingCS) {

				private IStorage storage;
				private boolean dirty = true;

				@Override
				protected IStatus run(final IProgressMonitor monitor) {
					final SubMonitor progress = SubMonitor.convert(monitor, 10);
					try {
						storage = ((IStorageEditorInput) input).getStorage();
					} catch (final CoreException e1) {
						return new CoreExceptionStatus(e1);
					}
					progress.worked(3);
					if (input.exists()) {
						final TextGridObject textGridObject = AdapterUtils.getAdapter(storage, TextGridObject.class);
						if (textGridObject != null && textGridObject.isNew()) {
							collationSet = new CollationSet();
							this.dirty = true;
							progress.worked(7);
						} else
							try {
								final InputStream contents = storage.getContents();
								progress.worked(3);
								collationSet = JAXB.unmarshal(contents, CollationSet.class);
								dirty = false;

								// now we're trying to complete all those TGOs within
								for (final CollationSetWitness witness : collationSet.getWitnesses())
									if (witness.getUri().getScheme().equals(TextGridObject.SCHEMA))
										TextGridObject.getInstance(
												witness.getUri(), true);

							} catch (final DataBindingException e) {
								return new Status(IStatus.WARNING, CollateXPlugin.PLUGIN_ID, NLS.bind(Messages.CollationSetEditor_FailedToParseX, storage), e);
							} catch (final CrudServiceException e) {
								// ignore this for now -- will be asked for
								// later anyway
							} catch (final CoreException e) {
								return new CoreExceptionStatus(
										CollateXPlugin.PLUGIN_ID, e);
							}
						progress.worked(4);
					}
					progress.done();
					return Status.OK_STATUS;
				}

				@Override
				protected IStatus done(final IJobChangeEvent event, final IProgressMonitor monitor) {

					if (!event.getResult().isOK())
						if (event.getResult().getException() != null && event.getResult().getException() instanceof DataBindingException)
							ErrorDialog.openError(getSite().getShell(), Messages.CollationSetEditor_ErrorReadingCS, NLS.bind(Messages.CollationSetEditor_ParseErrorExplanation, input.getName()), event.getResult(), 0xFFFF);
					setPartName(input.getName());
					readOnly = storage.isReadOnly();
					setDirty(dirty);
					setInputWithNotify(input);

					if (collationSet != null && collationSetViewer != null
							&& !collationSetViewer.getControl().isDisposed()) {
						collationSetViewer.setInput(collationSet);
						if (collationSet.getRegularisation() == null)
							regularisationLink
							.setText(Messages.CollationSetEditor_NewRegularisationLink);
						else
							regularisationLink
							.setText(NLS
									.bind(Messages.CollationSetEditor_RegularisationLink,
											TextGridObject
											.getInstanceOffline(collationSet
													.getRegularisation())));

					}

					return Status.OK_STATUS;

				}
			}.schedule();
		}
	}

	protected TextGridObject getInputObject() {
		if (getEditorInput() instanceof IStorageEditorInput) {
			final IStorageEditorInput input = (IStorageEditorInput) getEditorInput();
			try {
				return AdapterUtils.getAdapter(input.getStorage(),
						TextGridObject.class);
			} catch (final CoreException e) {
				StatusManager.getManager().handle(e, CollateXPlugin.PLUGIN_ID);
			}
		}
		return null;

	}

	@Override
	public boolean isDirty() {
		return dirty;
	}

	public void setDirty(final boolean dirty) {
		this.dirty = dirty;
		firePropertyChange(PROP_DIRTY);
	}

	@Override
	public boolean isSaveAsAllowed() {
		return false;
	}

	protected CheckboxTreeViewer getCollationSetViewer() {
		return collationSetViewer;
	}

	/**
	 * Triggers the actual collation (after some optional confirmations by the
	 * user)
	 */
	public void collate() {

		final SaveConfirmation saveDialog = new SaveConfirmation(
				getSite().getShell(),
				Messages.CollationSetEditor_SaveConfirmationMessage);
		saveDialog.addFilter(new Predicate<IEditorReference>() {

			@Override
			public boolean apply(final IEditorReference ref) {
				IEditorInput input;
				try {
					input = ref.getEditorInput();
					if (input.equals(getEditorInput()))
						return true;
					final TextGridObject tgo = AdapterUtils.getAdapter(input,
							TextGridObject.class);
					if (collationSet.getRegularisation() != null)
						try {
							final TextGridObject reg = TextGridObject
									.getInstanceOffline(collationSet
											.getRegularisation());
							if (reg.equals(tgo))
								return true;
						} catch (final AssertionFailedException e) {
							// TODO handle non-tgo stuff?
						}
				} catch (final PartInitException e1) {
					StatusManager.getManager().handle(e1,
							CollateXPlugin.PLUGIN_ID);
				}
				return false;
			}

		});
		if (!saveDialog.ask())
			return;

		if (collateJob != null && collateJob.getState() == Job.RUNNING) {
			final MessageDialog dialog = new MessageDialog(
					getSite().getShell(),
					Messages.CollationSetEditor_CollationRunningTitle,
					null,
					Messages.CollationSetEditor_CollationRunningMessage,
					MessageDialog.QUESTION, new String[] { Messages.CollationSetEditor_NewCollation,
						Messages.CollationSetEditor_WaitForCollation }, 0);
			dialog.setBlockOnOpen(true);
			final int choice = dialog.open();
			if (choice == 0)
				collateJob.cancel();
			else
				return;
		}

		collateJob = new CollateJob(NLS.bind(Messages.CollationSetEditor_CollateJobTitle, getEditorInput()
				.getName()), collationSet);
		collateJob.addJobChangeListener(new JobChangeAdapter() {

			@Override
			public void done(final IJobChangeEvent event) {
				if (event.getResult().isOK())
					setCollationResults(((CollateJob) event.getJob())
							.getCollationResults());
			}

		});
		collateJob.schedule();
	}

	public CollationSetWitness[] getSelectedWitnesses() {
		if (collationSetViewer != null) {
			final IStructuredSelection sel = (IStructuredSelection) collationSetViewer
					.getSelection();
			return AdapterUtils.getAdapters(sel.toArray(),
					CollationSetWitness.class, false);
		}
		return new CollationSetWitness[0];
	}

	/**
	 * Deletes all currently selected witnesses.
	 */
	public void deleteSelection() {
		final CollationSetWitness[] witnesses = getSelectedWitnesses();
		getCollationSet().getWitnesses().removeAll(
				ImmutableList.copyOf(witnesses));
		getCollationSetViewer().remove(getCollationSet(), witnesses);
		setDirty(true);
	}

	@Override
	public void resourceChanged(final IResourceChangeEvent event) {
		if (file != null) {
			final IResourceDelta delta = event.getDelta().findMember(
					file.getFullPath());
			if (delta != null)
				if ((delta.getFlags() & IResourceDelta.MOVED_TO) != 0) {
					final IFile newFile = (IFile) ResourcesPlugin
							.getWorkspace().getRoot()
							.findMember(delta.getMovedToPath());
					new UIJob(Messages.CollationSetEditor_AdaptingMove) {

						@Override
						public IStatus runInUIThread(
								final IProgressMonitor monitor) {
							setInput(new FileEditorInput(newFile));
							return Status.OK_STATUS;
						}

					}.schedule();
				}
		}
	}
}
